import os

from .base_service import BaseService
from servicer.git import Git

class BasePackageService(BaseService):
    def __init__(self, config=None):
        super().__init__(config=config)

        self.git = Git()

        if 'package_file_path' in self.config and 'version_file_path' not in self.config:
             self.config['version_file_path'] = self.config['package_file_path']

    def set_auto_version(self, max_increment=10):
        print('auto-versioning...')
        self.package_name = self.package_name(self.config['package_file_path'])
        self.package_version = self.package_version(self.config['version_file_path'])

        current_increment = 0
        version_changed = False
        while True:
            invalid_version = self.if_package_version_exists(package_name=self.package_name, version=self.package_version)

            if not invalid_version:
                break

            if current_increment > max_increment:
                raise ValueError('Max package_version auto-increment reached! %s-%s' % (self.package_name, self.package_version))

            self.package_version = self.increment_version(self.package_version)
            version_changed = True
            current_increment += 1

        print('Automatic version decided: %s-%s' % (self.package_name, self.package_version))

        if version_changed:
            self.write_package_version(path=self.config['version_file_path'], version=self.package_version)

    def commit_and_push_changes(self, protocol='ssh'):
        self.git.commit(add=self.config['version_file_path'], message='[servicer] Automated version change.')
        self.git.push(branch=self.git.current_branch(), protocol=protocol)

    def increment_version(self, version):
        new_version = [int(v) for v in version.split('.')]
        new_version[-1] += 1
        return '.'.join([str(v) for v in new_version])



from .base_auth_adapter import BaseAuthAdapter

import os
from servicer.run import run

class AuthAdapter(BaseAuthAdapter):
    def __init__(self, config):
        super().__init__(config)
        self.run = run

    def authenticate(self):
        self.run('gcloud -v')

        project = self.config.get('project', os.environ['PROJECT_NAME'])
        compute_zone = self.config.get('compute_zone', os.environ['GCLOUD_COMPUTE_ZONE'])
        key_file_path = self.config.get('key_file_path', os.environ['GCLOUD_KEY_FILE_PATH'])

        self.ensure_key_file(key_file_path)

        self.run('gcloud auth activate-service-account --key-file %s' % key_file_path)
        self.run('gcloud config set project %s' % project)
        self.run('gcloud config set compute/zone %s' % compute_zone)

    def current_user(self):
        result = self.run('gcloud auth list --filter=status:ACTIVE --format="value(account)"')
        active_users = result['stdout'].strip().split('\n')

        if active_users:
            return active_users[0]

    def ensure_key_file(self, key_file_path):
        if os.path.isfile(key_file_path):
            print('found existing gcloud key-file: %s' % key_file_path)
        else:
            print('generating gcloud key-file: %s' % key_file_path)
            key_file_json = self.config.get('key_file_json', os.environ['GCLOUD_KEY_FILE_JSON'])

            with open(key_file_path, 'w') as out:
                out.write(key_file_json)



import sys
import os
from urllib.parse import quote_plus

from .run import run

class Git():
    def __init__(self, hide_output=True):
        self.run = run
        self.hide_output = hide_output

    # deprecated, use diff with merge_base=True instead
    # def files_changed_ahead_of_ref(self, ref):
    #     forward_commits = self.commits_against_ref(ref, commit_type='+')
    #     files_changed = set()
    #     for c in forward_commits:
    #         files_changed.update(self.diff(a='%s~' % c, b=c, name_only=True))
    #     return sorted(list(files_changed))

    def authors_for_changes_ahead_of_ref(self, ref):
        forward_commits = self.commits_against_ref(ref, commit_type='+')
        authors = set()
        for c in forward_commits:
            authors.add(self.author_for_ref(c))
        return sorted(list(authors))

    def commits_against_ref(self, ref, commit_type='', hide_output=None):
        result = self.run('git cherry %s' % ref, hide_output=hide_output or self.hide_output)

        commits = result['stdout'].strip().split('\n')
        return [c[2:] for c in commits if c and c.startswith(commit_type)]

    def diff(self, a, b='HEAD', name_only=False, merge_base=False, hide_output=None):
        command = 'git diff %s' % a
        if merge_base:
            command = '%s...%s' % (command, b)
        else:
            command = '%s..%s' % (command, b)
        if name_only:
            command = '%s --name-only' % command

        result = self.run(command, hide_output=hide_output or self.hide_output)

        if name_only:
            return [f for f in result['stdout'].strip().split('\n') if len(f) > 0]

        return result['stdout'].strip()

    def tag(self, tag, message=None, push=False, hide_output=None):
        command = 'git tag'
        if message:
            command = '%s -a %s -m "%s"' % (command, tag, message)
        else:
            command = '%s %s' % (command, tag)

        self.run(command, hide_output=hide_output or self.hide_output)

        if push:
            self.run('git push origin %s' % tag)

    def list_tags(self):
        result = self.run('git tag')
        return result['stdout'].strip().split('\n')

    def sanitize_tag(self, tag):
        for ch in '/ [ ]'.split():
            tag.replace(ch, '.')
        return tag

    def set_config(self, config=None):
        for key, value in config.items():
            result = self.run('git config %s' % key)
            if not result['stdout'].strip():
                self.run('git config %s "%s"' % (key, value))

    def commit(self, add='.', message=None):
        if not isinstance(add, list):
            add = [add]

        for a in add:
            self.run('git add %s' % a)

        self.run('git diff-index --quiet HEAD || git commit -m "%s"' % message)

    def push(self, origin='origin', branch=None, protocol='ssh'):
        if protocol == 'ssh':
            self.run('git push %s %s' % (origin, branch))
        elif protocol == 'https':
            self.run('git push https://%s:%s@%s %s:%s' % (os.environ['GIT_USERNAME'], quote_plus(os.environ['GIT_PASSWORD']), os.environ['GIT_REPOSITORY'], branch, branch))
        else:
            raise ValueError('Invalid git push protocol: %s' % protocol)

    def current_branch(self):
        result = self.run('git rev-parse --abbrev-ref HEAD')
        return result['stdout'].strip()

    def author_for_ref(self, ref=None):
        command = 'git show --quiet --format="%an"'
        if ref:
            command = '%s %s' % (command, ref)

        result = self.run(command)
        return result['stdout'].strip()



from .base_ci_adapter import BaseCIAdapter

class CIAdapter(BaseCIAdapter):

    def __init__(self):
        super().__init__()

        # https://confluence.atlassian.com/bitbucket/environment-variables-794502608.html
        self.env_map['BITBUCKET_BUILD_NUMBER'] = 'BUILD_NUMBER'
        self.env_map['BITBUCKET_CLONE_DIR'] = 'CLONE_DIR'
        self.env_map['BITBUCKET_COMMIT'] = 'COMMIT'
        self.env_map['BITBUCKET_REPO_OWNER'] = 'REPO_OWNER'
        self.env_map['BITBUCKET_REPO_OWNER_UUID'] = 'REPO_OWNER_UUID'
        self.env_map['BITBUCKET_REPO_SLUG'] = 'REPO_SLUG'
        self.env_map['BITBUCKET_REPO_UUID'] = 'REPO_UUID'
        self.env_map['BITBUCKET_BRANCH'] = 'BRANCH'
        self.env_map['BITBUCKET_TAG'] = 'TAG'
        self.env_map['BITBUCKET_BOOKMARK'] = 'BOOKMARK'
        self.env_map['BITBUCKET_PARALLEL_STEP'] = 'PARALLEL_STEP'
        self.env_map['BITBUCKET_PARALLEL_STEP_COUNT'] = 'PARALLEL_STEP_COUNT'



import os
import sys
import hashlib
import requests
from requests.auth import HTTPBasicAuth
from ..base_service import BaseService

class Service(BaseService):
    def __init__(self, config):
        super().__init__(config)

        self.endpoint = os.environ['ARTIFACTORY_ENDPOINT']
        self.username = os.environ['ARTIFACTORY_USERNAME']
        self.password = os.environ['ARTIFACTORY_PASSWORD']

    def up(self):
        super().up()

        if 'steps' in self.config:
            self.run_steps(self.config['steps'])

    def down(self):
        super().up()

    def run_steps(self, steps):
        for step in steps:
            getattr(self, step['type'])(**step['args'])

    def upload_files(self, artifactory_directory=None, local_paths=None, errorOnExists=None):
        paths = []
        for lp in local_paths:
            if os.path.isdir(lp):
                for dir_path in os.listdir(lp):
                    paths.append({
                        'ap': '%s/%s' % (artifactory_directory, dir_path),
                        'lp': '%s/%s' % (lp, dir_path),
                    })
            else:
                paths.append({'ap':'%s/%s' % (artifactory_directory, lp), 'lp':lp})

        for path in paths:
            self.upload_file(artifactory_path=path['ap'], local_path=path['lp'], errorOnExists=errorOnExists)

    def upload_file(self, artifactory_path=None, local_path=None, properties=None, errorOnExists=None):
        print('uploading: %s -> %s' % (local_path, artifactory_path))

        if errorOnExists:
            self.file_exists(artifactory_path=artifactory_path, action='error')

        buffer_size = 65536  # read file in 64kb chunks

        md5 = hashlib.md5()
        sha1 = hashlib.sha1()

        with open(local_path, 'rb', buffering=0) as f:
            for block in iter(lambda : f.read(buffer_size), b''):
              md5.update(block)
              sha1.update(block)

        print("MD5: {0}".format(md5.hexdigest()))
        print("SHA1: {0}".format(sha1.hexdigest()))

        url = '%s/%s' % (self.endpoint, artifactory_path)

        if properties:
            for key, value in properties.items():
                url = '%s;%s=%s' % (url, key, value)

        # requests.exceptions.HTTPError: 409 Client Error: Conflict for url ???
        # files = {'file': open(local_path, 'rb')}
        # response = requests.put(
        #     url,
        #     auth=HTTPBasicAuth(self.username, self.password),
        #     headers={'X-Checksum-Sha1':sha1.hexdigest()},
        #     files=files,
        # )
        # response.raise_for_status()

        # workaround for 409 error
        self.run('curl -i -u "%s:%s" -H "X-Checksum-Sha1:%s" -T %s "%s"' % (self.username, self.password, sha1.hexdigest(), local_path, url))

    def download_files(self, artifactory_paths=None, local_directory=None):
        for ap in artifactory_paths:
            file_name = ap.split('/')[-1]
            local_path = '%s/%s' % (local_directory, file_name)
            self.download_file(ap, local_path)

    def download_file(self, artifactory_path=None, local_path=None):
        print('downloading: %s -> %s' % (artifactory_path, local_path))

        url = '%s/%s' % (self.endpoint, artifactory_path)
        response = requests.get(url, auth=HTTPBasicAuth(self.username, self.password), stream=True)
        response.raise_for_status()

        os.makedirs(os.path.dirname(local_path), exist_ok=True)
        with open(local_path, 'wb') as handle:
            for block in response.iter_content(1024):
                handle.write(block)

    def file_exists(self, artifactory_path=None, action=None):
        print('searching for file: %s' % artifactory_path)
        file_name = artifactory_path.split('/')[-1]
        url = '%s/api/search/artifact?name=%s' % (self.endpoint, file_name)

        response = requests.get(url, auth=HTTPBasicAuth(self.username, self.password))
        response.raise_for_status()

        body = response.json()

        exists = False
        for result in body['results']:
            print('result: %s' % result)
            if result['uri'].endswith(artifactory_path):
                exists = True

        if exists and action == 'error':
            raise ValueError('Artifact already exists! %s' % artifactory_path)

        return exists



from unittest import TestCase, mock

from servicer.servicer import Servicer

class ServicerTest(TestCase):
    def setUp(self):
        self.servicer = Servicer(args={})

    # def AutoMock(self, **attributes):
    #     automock = Mock(spec=attributes.keys())
    #     for k, v in attributes.items():
    #         setattr(automock, k, v)
    #     return automock

class ServicerClassTest(ServicerTest):
    def test_initialized(self):
        pass

class GetServiceEnvironmentTest(ServicerTest):
    def setUp(self):
        super().setUp()

        self.servicer.map_service_environment = mock.Mock(return_value=None)

    def test_defaults(self):
        result = self.servicer.get_service_environment('my-branch')
        self.servicer.map_service_environment.assert_called_with('my-branch', [
            { 'branch': 'master', 'environment': 'production' },
            { 'branch': 'develop' },
            { 'branch': 'env-*' },
        ])
        self.assertEqual(result, None)

    def test_custom_mappings(self):
        self.servicer.config['environment'] = {
            'mappings': [{ 'branch': 'foo', 'environment': 'bar' }],
        }
        result = self.servicer.get_service_environment('my-branch')
        self.servicer.map_service_environment.assert_called_with('my-branch', self.servicer.config['environment']['mappings'])
        self.assertEqual(result, None)

    def test_matched_environment(self):
        self.servicer.map_service_environment.return_value = 'my_wacky/environment'

        result = self.servicer.get_service_environment('my-branch')
        self.servicer.map_service_environment.assert_called_with('my-branch', mock.ANY)
        self.assertEqual(result, 'my-wacky-environment')

class MapServiceEnvironmentTest(ServicerTest):
    def test_no_mappings(self):
        result = self.servicer.map_service_environment('my-branch', [])
        self.assertEqual(result, None)

    def test_branch_mapping(self):
        result = self.servicer.map_service_environment('develop', [
            { 'branch': 'develop' },
        ])
        self.assertEqual(result, 'develop')

    def test_branch_environment_mapping(self):
        result = self.servicer.map_service_environment('master', [
            { 'branch': 'master', 'environment': 'production' },
        ])
        self.assertEqual(result, 'production')

    def test_branch_wildcard_mapping(self):
        result = self.servicer.map_service_environment('env-my-branch', [
            { 'branch': 'env-*' },
        ])
        self.assertEqual(result, 'env-my-branch')

    def test_regex_mapping(self):
        result = self.servicer.map_service_environment('env-my-branch', [
            { 'branch': '/.*-my-.*/' },
        ])
        self.assertEqual(result, 'env-my-branch')

    def test_no_match(self):
        result = self.servicer.map_service_environment('my-branch', [
            { 'branch': 'master', 'environment': 'production' },
            { 'branch': 'develop' },
            { 'branch': 'env-*' },
        ])
        self.assertEqual(result, None)

    def test_multiple_match(self):
        result = self.servicer.map_service_environment('env-my-branch-qa', [
            { 'branch': 'master', 'environment': 'production' },
            { 'branch': '*-qa', 'environment': 'qa' },
            { 'branch': 'env-*' },
        ])
        self.assertEqual(result, 'qa')



from ..base_service import BaseService

class GCloudService(BaseService):
    def __init__(self, config):
        super().__init__(config)
        self.name = 'a-gcloud-service'

    def up(self):
        super().up()

    def down(self):
        super().down()



import re

def interpolate_tokens(config, params, ignore_missing_key=False):
    if isinstance(config, dict):
        for key, value in config.items():
            if isinstance(value, str):
                config[key] = replace_tokens(value, params, ignore_missing_key)
            else:
                interpolate_tokens(value, params, ignore_missing_key)
    elif isinstance(config, list):
        for i in range(len(config)):
            if isinstance(config[i], str):
                config[i] = replace_tokens(config[i], params, ignore_missing_key)
            else:
                interpolate_tokens(config[i], params, ignore_missing_key)

def replace_tokens(value, params, ignore_missing_key=False):
    for match in re.findall(r'\${.+?}', value):
        key = match[2:-1]

        if ignore_missing_key and key not in params:
            continue

        value = re.sub(r'\${%s}' % key, params[key], value)

    return value



from ..base_service import BaseService

class AWSService(BaseService):
    def __init__(self, config):
        print(config)
        super().__init__(config)
        self.name = 'an-aws-service'

        import boto3
        self.boto3 = boto3

    def up(self):
        super().up()

    def down(self):
        super().down()



import os
import sys
import importlib
import json
import argparse
import yaml
import re
import copy

from .generate_ci_config import generate_ci_config
from .topological_order import toposort2
from .run import run
from .git import Git
from .config_loader import ConfigLoader

class Servicer():
    def __init__(self, args=None):
        self.run = run
        self.git = Git(hide_output=('DEBUG' not in os.environ))

        if args == None:
            args = vars(self.load_arguments())

        self.load_environment(args)
        self.config_loader = ConfigLoader(args)
        self.config = self.config_loader.load_config()

        self.normalize_ci_environment()
        self.git_init()

        self.services = self.load_service_modules()
        self.service_order = self.order_services(self.services)

        if 'generate_ci' in self.config['args'] and self.config['args']['generate_ci']:
            generate_ci_config(config=config, path='.')
            sys.exit(0)

        if 'step' in self.config['args']:
            self.config['args']['step'] = self.config['args']['step'].split(',')

    def load_arguments(self):
        parser = argparse.ArgumentParser(description='Process deployment options.')
        parser.add_argument('--generate_ci', action='store_true', help='generate a ci config file, do not run any deploy options')
        parser.add_argument('--service', help='deploy only the provided service')
        parser.add_argument('--services_file', default='services.yaml', help='custom path to your services config file (default is services.yaml)')
        parser.add_argument('--servicer_config_path', default=f'{os.getcwd()}/.servicer', help='path to your servicer directory (default is ./servicer)')
        parser.add_argument('--step', default='deploy', help='perform the comma-separated build steps, defaults to deploy only')
        return parser.parse_args()

    def load_environment(self, args):
        if 'servicer_config_path' not in args:
            return

        env_file = '%s/.env.yaml' % args['servicer_config_path']

        print('checking for (.env.yaml) at (%s)' % env_file)
        if os.path.exists(env_file):

            print('(.env.yaml) found, including these arguments:')

            yaml_dict = yaml.load(open(env_file))
            for key, value in yaml_dict.items():
                os.environ[key] = value
                print(key)
            print()

    def normalize_ci_environment(self):
        if 'ci' not in self.config:
            return

        self.print_title('normalizing CI environment')

        self.config['ci']['adapters'] = {}
        for p in self.config['ci']['providers']:
            print('CI Adapter: %s' % p)
            ci_adapter_modules = [
                {
                    'name': 'ci_adapters.%s' % p,
                    'package': 'ci_adapters',
                    'file_path': '%s/ci_adapters/%s.py' % (self.config['config_path'], p),
                },
                {
                    'name': 'servicer.builtin.ci_adapters.%s' % p,
                    'package': 'servicer.builtin.ci_adapters',
                    'file_path': '%s/builtin/ci_adapters/%s.py' % (self.config['module_path'], p),
                },
            ]
            module = self.load_module_from_paths(ci_adapter_modules)
            self.config['ci']['adapters'][p] = module.CIAdapter()

        for ci_adapter in self.config['ci']['adapters'].values():
            ci_adapter.convert_environment_variables()

        print()
        self.service_environment = os.getenv('SERVICE_ENVIRONMENT') or self.get_service_environment(os.environ['BRANCH'])

        print('branch: %s' % os.getenv('BRANCH'))
        print('service environment: %s' % self.service_environment)
        if self.service_environment:
            os.environ['SERVICE_ENVIRONMENT'] = self.service_environment

    def git_init(self):
        if not 'git' in self.config or not self.config['git']['enabled']:
            return

        self.print_title('initializing git integration')

        # if 'config' in self.config['git']:
        #     self.git.set_config(self.config['git']['config'])

        if 'GIT_DIFF_REF' in os.environ:
            result = self.run('git cat-file -t %s' % os.environ['GIT_DIFF_REF'], check=False)
            if result['status'] != 0:
                print('Invalid GIT_DIFF_REF provided!')
            else:
                self.config['git']['diff-ref'] = os.environ['GIT_DIFF_REF']

        if 'diff-ref' not in self.config['git'] and 'BRANCH' in os.environ and 'default-branch' in self.config['git']:
            if os.environ['BRANCH'] != self.config['git']['default-branch']:
                print('defaulting Git Diff Ref to default-branch')
                self.config['git']['diff-ref'] = 'origin/%s' % self.config['git']['default-branch']

        if 'diff-ref' in self.config['git']:
            print('Git Diff Ref: %s\n' % self.config['git']['diff-ref'])

        if self.config['git']['ignore-servicer-commits'] and 'diff-ref' in self.config['git']:
            authors = self.git.authors_for_changes_ahead_of_ref(self.config['git']['diff-ref'])
            print('Commit authors: %s' % authors)
            if authors == ['servicer']:
                print('Only automated servicer changes were detected, skipping this build.')
                sys.exit(0)

    def load_service_modules(self):
        self.print_title('loading service modules')

        if 'services' in self.config:
            for name, service in self.config['services'].items():
                service['name'] = name

        services = {}
        if 'service' in self.config['args'] and self.config['args']['service']:
            service_names = self.config['args']['service'].split(',')
            for name in service_names:
                services[name] = self.config['services'][name]
        elif 'services' in self.config:
            services = self.config['services']
            self.ignore_unchanged_services(services)

        for service_name, service in services.items():
            if 'config' not in service:
                service['config'] = {}

            self.try_initialize_provider(service['provider'], service)

            service['name'] = service_name
            service_modules = [
                {
                    'name': 'service_adapters.%s.%s' % (service['provider'], service['service_type']),
                    'package': 'service_adapters',
                    'file_path': '%s/service_adapters/%s/%s.py' % (self.config['config_path'], service['provider'], service['service_type']),
                },
                {
                    'file_path': '%s/service_adapters/%s/%s.sh' % (self.config['config_path'], service['provider'], service['service_type']),
                },
                {
                    'name': 'servicer.builtin.service_adapters.%s.%s' % (service['provider'], service['service_type']),
                    'package': 'servicer.builtin.service_adapters',
                    'file_path': '%s/builtin/service_adapters/%s/%s.py' % (self.config['module_path'], service['provider'], service['service_type']),
                },
                {
                    'file_path': '%s/builtin/service_adapters/%s/%s.sh' % (self.config['module_path'], service['provider'], service['service_type']),
                },
            ]
            module = self.load_module_from_paths(service_modules)
            if isinstance(module, str):
                service['shell_script'] = service['module']
            else:
                service['module'] = module

        return services

    def ignore_unchanged_services(self, services):
        if not 'git' in self.config or not self.config['git']['enabled'] or not self.config['git']['ignore-unchanged']:
            return

        if 'diff-ref' not in self.config['git']:
            print('No GIT_DIFF_REF found, aborting change detection.')
            return

        if 'BRANCH' in os.environ:
            sanitized_tag = 'servicer-%s' % self.git.sanitize_tag(os.environ['BRANCH'])
            tags = [t for t in self.git.list_tags() if t.startswith(sanitized_tag)]

            if os.getenv('DEBUG'):
                print('branch tag: %s' % sanitized_tag)
                print('matching tags:')
                print('\n'.join(tags))

            if len(tags) > 0:
                git_ref = tags[-1]

        diff_files = self.git.diff(self.config['git']['diff-ref'], name_only=True, merge_base=True)
        print('\nChanged Files:')
        print('\n'.join(diff_files))

        # TODO: think through what top level 'watch_paths' means
        if 'ignore_paths' in self.config['git']:
            regexes = [self.sanitize_regex(matcher) for matcher in self.config['git']['ignore_paths']]
            matched_files, diff_files = self.match_regexes(diff_files, regexes)

        ignored_services = []
        for service_name, service in services.items():
            service_changed_files = diff_files

            if 'git' in service:
                if 'watch_paths' in service['git']:
                    watch_regexes = [self.sanitize_regex(matcher) for matcher in service['git']['watch_paths']]

                    if os.getenv('DEBUG'):
                        print('\nService: %s' % service_name)
                        print('Matchers:')

                    service_changed_files, _ = self.match_regexes(diff_files, watch_regexes)

                if 'ignore_paths' in service['git']:
                    ignore_regexes = [self.sanitize_regex(matcher) for matcher in service['git']['ignore_paths']]
                    _, service_changed_files = self.match_regexes(service_changed_files, ignore_regexes)

                if len(service_changed_files) > 0:
                    if os.getenv('DEBUG'):
                        print('\nChanged Files:')
                        print('\n'.join(service_changed_files))
                else:
                    ignored_services.append(service_name)

        print('\nIgnored Services:')
        for sn in ignored_services:
            print(sn)
            services.pop(sn)

        print('\nChanged Services:')
        print('\n'.join(services.keys()))

    def sanitize_regex(self, matcher):
        if matcher.startswith('/') and matcher.endswith('/'):
            return matcher[1:-1]

        return matcher.replace('*', '.*')

    # takes a list of strings, and returns a tuple of strings that match and do not match
    def match_regexes(self, strings, regexes):
        if not isinstance(regexes, list):
            regexes = [regexes]

        matches = []
        unmatches = []
        for s in strings:
            for regex in regexes:
                result = re.match(regex, s)

                if os.getenv('DEBUG'):
                    print('%s ~= %s -> %s' % (regex, s, result))

                if result:
                    matches.append(s)
                else:
                    unmatches.append(s)

        return matches, unmatches

    def try_initialize_provider(self, provider_name, service):
        provider = self.config['providers'].get(provider_name)

        if provider == None:
            return

        provider['name'] = provider_name

        should_initialize = self.service_environment and not provider.get('initialized')
        force_authenticate = False

        provider_config = copy.deepcopy(provider)

        if service and 'auth' in service:
            self.config_loader.merge_config(provider_config, service['auth'])
            should_initialize = True
            force_authenticate = True

        if should_initialize:
            self.initialize_provider(provider_config, force_authenticate=force_authenticate)

        if service:
            service['config']['initialized_provider'] = provider_config

    def initialize_provider(self, provider, force_authenticate=False):
        self.print_title('intializing provider: %s' % provider['name'])
        print('force_authenticate: %s' % force_authenticate)

        if 'libraries' in provider and provider['libraries']:
            self.run('%s install %s' % (os.getenv('PIP_EXE', 'pip'), ' '.join(provider['libraries'])))
        if 'auth_script' in provider and provider['auth_script']:
            print('authenticating with: %s' % provider['name'])
            auth_script_paths = [
                { 'file_path': '%s/auth/%s.sh' % (self.config['config_path'], provider['name']) },
                { 'file_path': '%s/builtin/auth/%s.sh' % (self.config['module_path'], provider['name']) },
            ]
            auth_script_path = self.load_module_from_paths(auth_script_paths)
            provider['auth_script_path'] = auth_script_path
            self.run(auth_script_path)
        if 'config' in provider:
            auth_modules = [
                {
                    'name': 'auth_adapters.%s' % provider['name'],
                    'package': 'auth_adapters',
                    'file_path': '%s/auth_adapters/%s.py' % (self.config['config_path'], provider['name']),
                },
                {
                    'name': 'servicer.builtin.auth_adapters.%s' % provider['name'],
                    'package': 'servicer.builtin.auth_adapters',
                    'file_path': '%s/builtin/auth_adapters/%s.py' % (self.config['module_path'], provider['name']),
                },
            ]
            module = self.load_module_from_paths(auth_modules)
            auth = module.AuthAdapter(provider['config'])

            if force_authenticate or not auth.current_user():
                auth.authenticate()

            print('Current User: %s' % auth.current_user())

        provider['initialized'] = True

    def load_module_from_paths(self, modules):
        for mp in modules:
            if os.getenv('DEBUG'):
                print('searching for module at: %s' % mp['file_path'])
            if os.path.exists(mp['file_path']):
                if 'name' in mp:
                    if os.getenv('DEBUG'):
                        print('importing: %s:%s' % (mp['name'], mp['package']))

                    module = importlib.import_module(mp['name'])
                    return module
                else:
                    print('found matching executable: %s' % mp['file_path'])
                    return mp['file_path']

        print('no module found!')
        sys.exit(1)

    def order_services(self, services):
        service_names = services.keys()
        dependencies = {}
        for service_name, service in services.items():
            if 'depends_on' in service:
                dependencies[service_name] = set(service['depends_on'])
            else:
                dependencies[service_name] = set()

        ordered_services = toposort2(dependencies)
        return [item for sublist in ordered_services for item in sublist]

    def run_steps(self):
        for step in self.config['args']['step']:
            self.print_title('%s step' % step)

            if step == 'deploy':
                if self.service_environment:
                    self.up()
                else:
                    print('skipping, no valid deploy environment was matched')
            else:
                for service_name in self.service_order:
                    service = self.services[service_name]

                    if 'steps' in service and step in service['steps']:
                        commands = service['steps'][step].get('commands')
                        if commands:
                            for c in commands:
                                self.run(c)

                        if 'module' in service and 'config' in service['steps'][step]:
                            config = service['steps'][step].get('config')
                            print('Step Config (%s): ' % step)
                            print(json.dumps(config, indent=4, sort_keys=True))
                            adapter = service['module'].Service(config)
                            adapter.up()

                        post_commands = service['steps'][step].get('post_commands')
                        if post_commands:
                            for c in post_commands:
                                self.run(c)

    def get_service_environment(self, branch):
        mappings = None
        if 'environment' in self.config and 'mappings' in self.config['environment']:
            mappings = self.config['environment']['mappings']
        else:
            mappings = [
                { 'branch': 'master', 'environment': 'production' },
                { 'branch': 'develop' },
                { 'branch': 'env-*' },
            ]

        service_environment = self.map_service_environment(branch, mappings)

        if service_environment:
            for ch in ['\\', '/', '_']:
                if ch in service_environment:
                    service_environment = service_environment.replace(ch, '-')
            return service_environment

    def map_service_environment(self, branch, mappings=[]):
        for m in mappings:
            if 'branch' in m:
                if m['branch'].startswith('/') and m['branch'].endswith('/'):
                    regex = m['branch'][1:-1]
                else:
                    regex = m['branch'].replace('*', '.*')

                result = re.match(regex, branch)
                if result:
                    return m.get('environment', branch)
            elif 'tag' in m:
                # TODO: support tag mapping
                pass

        return None

    def up(self):
        for service_name in self.service_order:
            print(service_name)

        for service_name in self.service_order:
            self.up_service(self.config['services'][service_name])

        self.print_title('deploy complete')

    def up_service(self, service):
        self.print_title('deploy %s' % service['name'])

        service['results'] = {}

        if 'module' in service:
            adapter = service['module'].Service(config=service['config'])
            results = adapter.up()
            if results:
                self.store_results(service, results)
        elif 'shell_script' in service:
            self.run('%s up' % service['shell_script'])

            # clunky way to retrieve results from shell scripts
            script_results_path = 'shell-script-results.yaml'
            if os.path.exists(script_results_path):
                print('loading script results from (%s)' % script_results_path)
                results = yaml.load(open(script_results_path))
                self.store_results(service, results)
                os.remove(script_results_path)

    def store_results(self, service, results):
        for key, value in results.items():
            if type(value) == str:
                os.environ[key.upper()] = value
        service['results']['up'] = results

    def print_title(self, message='', border='----'):
        inner_text = ' %s ' % message
        border_text = border

        while len(border_text) < len(inner_text):
            border_text = '%s %s' % (border_text, border)

        print()
        print(border_text)
        print(inner_text)
        print(border_text)
        print()

def main():
    Servicer().run_steps()

if __name__ == '__main__':
    main()



import os

from servicer.tokens import interpolate_tokens

class BaseAuthAdapter():

    def __init__(self, config):
        self.config = config
        interpolate_tokens(self.config, os.environ)

    def authenticate(self):
        pass

    def current_user(self):
        pass



from .base_ci_adapter import BaseCIAdapter

class CIAdapter(BaseCIAdapter):

    def __init__(self):
        super().__init__()

        # https://wiki.jenkins.io/display/JENKINS/Building+a+software+project
        self.env_map['BRANCH_NAME'] = 'BRANCH'
        self.env_map['BUILD_DISPLAY_NAME'] = 'BUILD_NAME'
        self.env_map['BUILD_ID'] = 'BUILDID'
        self.env_map['BUILD_NUMBER'] = 'BUILD_NUMBER'
        self.env_map['BUILD_TAG'] = 'TAG'
        self.env_map['BUILD_URL'] = 'BUILD_URL'
        self.env_map['CHANGE_AUTHOR'] = 'PR_USERNAME'
        self.env_map['CHANGE_AUTHOR_DISPLAY_NAME'] = 'PR_AUTHOR_NAME'
        self.env_map['CHANGE_AUTHOR_EMAIL'] = 'PR_AUTHOR_EMAIL'
        self.env_map['CHANGE_ID'] = 'PR_NUMBER'
        self.env_map['CHANGE_TARGET'] = 'PR_TARGET'
        self.env_map['CHANGE_TITLE'] = 'PR_TITLE'
        self.env_map['CHANGE_URL'] = 'PR_URL'
        self.env_map['EXECUTOR_NUMBER'] = 'NODE_INDEX'
        self.env_map['GIT_BRANCH'] = 'BRANCH'
        self.env_map['GIT_COMMIT'] = 'COMMIT'
        self.env_map['GIT_URL'] = 'REPO_URL'
        self.env_map['JENKINS_HOME'] = 'HOME'
        self.env_map['JENKINS_URL'] = 'CI_URL'
        self.env_map['JOB_BASE_NAME'] = 'JOB_BASE_NAME'
        self.env_map['JOB_NAME'] = 'JOB_NAME'
        self.env_map['JOB_URL'] = 'JOB_URL'
        self.env_map['NODE_LABELS'] = 'NODE_LABELS'
        self.env_map['NODE_NAME'] = 'NODE_NAME'
        self.env_map['WORKSPACE'] = 'WORKING_DIRECTORY'



from ..base_docker_image import DockerImageService

class Service(DockerImageService):
    def __init__(self, config):
        super().__init__(config)
        self.registry = 'us.gcr.io'

    # def run(self, command):
    #     auth_script = self.config['initialized_provider']['auth_script_path']
    #     command = '%s && %s' % (auth_script, command)
    #     super().run(command, shell=False)



from .aws_service import AWSService
import os
import requests
from botocore.client import ClientError

class Service(AWSService):
    def up(self):
        super().up()

        self.s3 = self.boto3.resource('s3')

        self.ensure_bucket()

        if 'steps' in self.config:
            self.run_steps(self.config['steps'])

    def down(self):
        super().down()

    def ensure_bucket(self):
        bucket_args = {
            'Bucket': self.config['bucket'],
        }

        print('ensuring bucket exists: %s' % bucket_args['Bucket'])
        try:
            self.s3.meta.client.head_bucket(**bucket_args)
        except ClientError:
            bucket_args['CreateBucketConfiguration'] = {
                'LocationConstraint': self.config['region'],
            }
            bucket = self.s3.create_bucket(**bucket_args)

    def run_steps(self, steps):
        for step in steps:
            getattr(self, step['type'])(**step['args'])

    def sync(self, from_path=None, to_path=None, delete=True):
        command = f'aws s3 sync {from_path} {to_path}'
        if delete:
            command = f'{command} --delete'
        self.run(command, shell=True)

    def url_to_s3(self, key=None, url=None):
        print('downloading: %s -> s3://%s/%s' % (url, self.config['bucket'], key))
        req_for_image = requests.get(url, stream=True)
        file_object_from_req = req_for_image.raw
        req_data = file_object_from_req.read()

        self.s3.Bucket(self.config['bucket']).put_object(Key=key, Body=req_data)
